# author:    Kieran McVeigh
# website:   http://www.bklynhlth.com
import numpy as np
import pandas as pd
import os
import json
import logging

import cv2
from deepface import DeepFace
from dataclasses import dataclass, field
from sklearn.cluster import KMeans

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

@dataclass
class FaceData:
    face: np.ndarray = field(default_factory=lambda: np.array([]), repr=False)
    face_embedding: list = field(default_factory=list)
    bb_x: int = 0
    bb_y: int = 0
    bb_w: int = 0
    bb_h: int = 0
    right_eye_x: int = 0
    right_eye_y: int = 0
    left_eye_x: int = 0
    left_eye_y: int = 0
    confidence: float = 0.0
    frame_idx: int = -1

    def __post_init__(self):
        if not isinstance(self.face, np.ndarray):
            raise ValueError("Face must be a numpy ndarray.")

    def to_dict(self, keep_face_image=False):
        """
        ---------------------------------------------------------------------------------------------------

        This function converts the FaceData object to a dictionary.

        Parameters:
        ............
        keep_face_image : bool
            Whether to include the face image in the dictionary.

        Returns:
        ............
        facedata_dict : dict
            Dictionary representation of the FaceData object.
        """
        facedata_dict = {
            'face': self.face if keep_face_image else np.array([]),
            'face_embedding': self.face_embedding,
            'bb_x': self.bb_x,
            'bb_y': self.bb_y,
            'bb_w': self.bb_w,
            'bb_h': self.bb_h,
            'right_eye_x': self.right_eye_x,
            'right_eye_y': self.right_eye_y,
            'left_eye_x': self.left_eye_x,
            'left_eye_y': self.left_eye_y,
            'confidence': self.confidence,
            'frame_idx': self.frame_idx
        }
        return facedata_dict

    def set_face_from_image(self, image):
        """
        ---------------------------------------------------------------------------------------------------

        Crops the image using the stored bounding box coordinates and sets the cropped image as the face.
        Also checks to ensure the bounding box does not exceed the dimensions of the image.

        Parameters:
        ............
        image : np.ndarray
            A numpy array representing the entire image from which the face will be cropped.
        """
        if not isinstance(image, np.ndarray):
            raise ValueError("The provided image must be a numpy ndarray.")
        
        if (self.bb_x < 0 or self.bb_y < 0 or 
            self.bb_x + self.bb_w > image.shape[1] or 
            self.bb_y + self.bb_h > image.shape[0]):
            Warning("Bounding box exceeds the limits of the provided image. Setting bounding box to full frame and confidence to 0")

        cropped_image = image[self.bb_y:self.bb_y + self.bb_h, self.bb_x:self.bb_x + self.bb_w]
        self.face = cropped_image

def get_config(filepath, json_file):
    """
    ------------------------------------------------------------------------------------------------------

    This function reads the configuration file containing the column names for the output dataframes,
    and returns the contents of the file as a dictionary.

    Parameters:
    ...........
    filepath : str
        The path to the configuration file.
    json_file : str
        The name of the configuration file.

    Returns:
    ...........
    measures: A dictionary containing the names of the columns in the output dataframes.

    ------------------------------------------------------------------------------------------------------
    """
    dir_name = os.path.dirname(filepath)
    measure_path = os.path.abspath(os.path.join(dir_name, f"config/{json_file}"))

    file = open(measure_path)
    measures = json.load(file)
    return measures

def deepface_dict_to_facedata(data_dict, frame_idx):
    """
    ---------------------------------------------------------------------------------------------------

    Converts a dictionary with specific keys into a FaceData object.

    Parameters:
    ............
    data_dict : dict
        A dictionary containing keys for 'face', 'facial_area', and 'confidence'.
    frame_idx : int
        Frame index of the face data.

    Returns:
    ............
    face_data_instance : FaceData
        An instantiated object of FaceData filled with data from the dictionary.

    Raises:
    ............
    KeyError
        If a required key is missing in the dictionary.
    ValueError
        If data types are not as expected.
    ---------------------------------------------------------------------------------------------------
    """
    try:
        face = data_dict['face']
        bb_x = data_dict['facial_area']['x']
        bb_y = data_dict['facial_area']['y']
        bb_w = data_dict['facial_area']['w']
        bb_h = data_dict['facial_area']['h']
        right_eye_x = data_dict['facial_area']['right_eye'][0]
        right_eye_y = data_dict['facial_area']['right_eye'][1]
        left_eye_x = data_dict['facial_area']['left_eye'][0]
        left_eye_y = data_dict['facial_area']['left_eye'][1]
        confidence = data_dict['confidence']

        if not isinstance(face, np.ndarray):
            raise ValueError("The 'face' field must be a numpy ndarray.")

        face_data_instance = FaceData(
            face=face,
            bb_x=bb_x,
            bb_y=bb_y,
            bb_w=bb_w,
            bb_h=bb_h,
            right_eye_x=right_eye_x,
            right_eye_y=right_eye_y,
            left_eye_x=left_eye_x,
            left_eye_y=left_eye_y,
            confidence=confidence,
            frame_idx=frame_idx
        )
        return face_data_instance

    except KeyError as e:
        raise KeyError(f"Missing key in the input dictionary: {e}")
    except ValueError as e:
        raise ValueError(f"Value error: {e}")

def get_config(filepath, json_file):
    """
    ------------------------------------------------------------------------------------------------------

    This function reads the configuration file containing the column names for the output dataframes,
    and returns the contents of the file as a dictionary.

    Parameters:
    ...........
    filepath : str
        The path to the configuration file.
    json_file : str
        The name of the configuration file.

    Returns:
    ...........
    measures: A dictionary containing the names of the columns in the output dataframes.

    ------------------------------------------------------------------------------------------------------
    """
    dir_name = os.path.dirname(filepath)
    measure_path = os.path.abspath(os.path.join(dir_name, f"config/{json_file}"))

    file = open(measure_path)
    measures = json.load(file)
    return measures


def extract_face_rgb(frame, back_end_str, enforce_detection):
    """
    ---------------------------------------------------------------------------------------------------

    Extracts faces from a given frame in RGB format using the DeepFace library.

    Parameters:
    ............
    frame : np.ndarray
        The input frame in BGR format.
    back_end_str : str
        The backend string specifying the face detection model to use.
    enforce_detection : bool
        A boolean flag indicating whether to enforce face detection.

    Returns:
    ............
    faces_detected : list
        A list of dictionaries, where each dictionary represents a detected face.
    ---------------------------------------------------------------------------------------------------
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    faces_detected = DeepFace.extract_faces(
        frame,
        detector_backend=back_end_str,
        enforce_detection=enforce_detection,
        align=False
    )

    for face in faces_detected:
        face['face'] = np.array([])

    return faces_detected

def extract_face_bgr(frame, back_end_str, enforce_detection, align):
    """
    ---------------------------------------------------------------------------------------------------

    Extracts faces for backends with BGR format using the DeepFace library.

    Parameters:
    ............
    frame : np.ndarray
        The input frame in BGR format.
    back_end_str : str
        The backend string specifying the face detection model to use.
    enforce_detection : bool
        A boolean flag indicating whether to enforce face detection.
    align : bool
        A boolean flag indicating whether to align the face.

    Returns:
    ............
    faces_detected : list
        A list of dictionaries, where each dictionary represents a detected face.
    ---------------------------------------------------------------------------------------------------

    """
    faces_detected = DeepFace.extract_faces(
        frame,
        detector_backend=back_end_str,
        enforce_detection=enforce_detection,
        align=align
    )
    return faces_detected   

def extract_faces(
    frame, 
    frame_idx, 
    rgb_back_ends,
    back_end_str='mtcnn',
    align=True, 
    enforce_detection=False
):
    """
    ---------------------------------------------------------------------------------------------------

    Extracts faces from a given frame using the specified backend and alignment settings.

    Parameters:
    ............
    frame : np.ndarray
        The input frame in BGR format.
    frame_idx : int
        The frame index.
    back_end_str : str
        The backend string specifying the face detection model to use.
    align : bool
        A boolean flag indicating whether to align the face.
    enforce_detection : bool
        A boolean flag indicating whether to enforce face detection.

    Returns:
    ............
    faces_detected : list
        A list of FaceData objects representing detected faces.
    ---------------------------------------------------------------------------------------------------
    
    """
    if back_end_str in rgb_back_ends:
        faces_detected = extract_face_rgb(frame, back_end_str, enforce_detection)
    else:
        faces_detected = extract_face_bgr(frame, back_end_str, enforce_detection, align)

    faces_detected = [deepface_dict_to_facedata(face, frame_idx) for face in faces_detected]
    
    return faces_detected

def prep_face_data_for_embed(facedata_list, frame):
    """
    ---------------------------------------------------------------------------------------------------

    Ensures face data has all necessary components to embed face.

    Parameters:
    ............
    facedata_list : list
        List of FaceData objects.
    frame : np.ndarray
        The input frame in BGR format.

    Returns:
    ............
    facedata_list : list
        List of prepared FaceData objects.
    ---------------------------------------------------------------------------------------------------

    """
    for face_data in facedata_list:
        if face_data.face.shape[0] == 0:
            face_data.set_face_from_image(frame)
    return facedata_list

def embed_faces(face_data_list, model_name='Facenet'):
    """
    ---------------------------------------------------------------------------------------------------

    Embeds faces from a list of FaceData objects using the specified model.

    Parameters:
    ............
    face_data_list : list
        List of FaceData objects.
    model_name : str
        Name of the model to use for embedding.

    Returns:
    ............
    face_data_list : list
        List of FaceData objects with embeddings.
    ---------------------------------------------------------------------------------------------------
    
    """
    for face_data in face_data_list:
        deep_face_dict = DeepFace.represent(
            face_data.face,
            detector_backend='skip',
            model_name=model_name
        )
        face_data.face_embedding = deep_face_dict[0]['embedding']
        
        #save memory
        face_data.face = np.array([])

    return face_data_list 

def extract_embed_faces_from_frame(
    frame,
    frame_idx,
    rgb_back_ends,
    detector_backend='mtcnn',
    model_name='Facenet'
):
    """
    ---------------------------------------------------------------------------------------------------

    Extracts and embeds faces from a given frame using the specified backend and model.

    Parameters:
    ............
    frame : np.ndarray
        The input frame in BGR format.
    frame_idx : int
        The frame index.
    detector_backend : str
        The backend string specifying the face detection model to use.
    model_name : str
        Name of the model to use for embedding.

    Returns:
    ............
    face_data_list : list
        List of FaceData objects with embeddings.
    ---------------------------------------------------------------------------------------------------
    """
    try:
        face_data_list = extract_faces(
            frame,
            frame_idx,
            rgb_back_ends,
            back_end_str=detector_backend
        )

        face_data_list = prep_face_data_for_embed(face_data_list, frame)
        face_data_list = embed_faces(face_data_list, model_name=model_name)
    except Exception as e:
        logger.info(f"Error extracting and embedding faces: {e}, frame_idx: {frame_idx}")
        face_data_list = [FaceData()]
    return face_data_list

def load_facedata_from_video(
    video_path,
    rgb_back_ends,
    detector_backend='mtcnn',
    model_name='Facenet', 
    n_frames=np.inf, 
    capture_n_per_sec=3
):
    """
    ---------------------------------------------------------------------------------------------------

    Load video frames from a given video file, capturing only a specified
    number of frames per second.

    Parameters:
    ............
        video_path (str): Path to the video file.
        n_frames (int, optional): Maximum number of frames to capture. Defaults to np.inf.
        capture_n_per_sec (int, optional): Number of frames to capture per second. Defaults to 30.

    Returns:
    ............
        tuple: A tuple containing the list of captured frames, the list of their corresponding frame indices,
               and the frames per second (fps) of the video.
    ---------------------------------------------------------------------------------------------------
           
    """
    cap = cv2.VideoCapture(video_path)
    if not cap.isOpened():
        raise ValueError(f"Error opening video file: {video_path}")
    
    fps = cap.get(cv2.CAP_PROP_FPS)
    if capture_n_per_sec > fps:
        raise ValueError(f"Capture rate per second cannot exceed the video's frames per second, video fps: {fps}")

    num_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    n_skip = round(fps / capture_n_per_sec)
    facedata_list = []

    frame_index = 0
    n_frames_capture = 0

    while True:

        ret, frame = cap.read()
        if not ret:
            break

        if (frame_index % n_skip) == 0:
            frame_facedata = extract_embed_faces_from_frame(
                frame,
                frame_index, 
                rgb_back_ends,
                detector_backend=detector_backend,
                model_name=model_name
            )
            facedata_list.extend(frame_facedata)

            n_frames_capture += 1

        frame_index += 1

        if n_frames_capture >= n_frames:
            num_frames = n_frames * n_skip
            break

    cap.release()
    return facedata_list, fps, num_frames


def facedata_list_to_df(facedata_list):
    """
    ---------------------------------------------------------------------------------------------------

    Converts a list of FaceData objects to a pandas DataFrame.

    Parameters:
    ............
    facedata_list : list
        List of FaceData objects.

    Returns:
    ............
    facedata_df : pandas.DataFrame
        DataFrame containing the data from the FaceData objects.
    ---------------------------------------------------------------------------------------------------
    """
    facedata_dict_list = [facedata.to_dict() for facedata in facedata_list]
    facedata_df = pd.DataFrame(facedata_dict_list)
    facedata_df['cluster'] = -1
    return facedata_df

def cluster_embeddings(facedata_df, confidence_threshold=0.7, n_clusters=2):
    """
    ---------------------------------------------------------------------------------------------------
    This function clusters face embeddings based on confidence threshold and number of clusters.

    Parameters:
    ............
    facedata_df : pandas.DataFrame
        DataFrame containing face data.
    confidence_threshold : float
        Confidence threshold for face data.
    n_clusters : int
        Number of clusters to create.

    Returns:
    ............
    facedata_df : pandas.DataFrame
        DataFrame with cluster assignments for each face.
    ---------------------------------------------------------------------------------------------------
    
    """

    threshold_bools = facedata_df['confidence'] > confidence_threshold
    thresholded_facedata_df = facedata_df[threshold_bools]
    cluster_model = KMeans(n_clusters=n_clusters)
    cluster_assignments = cluster_model.fit_predict(
        np.vstack(thresholded_facedata_df['face_embedding'])
    )
    facedata_df.loc[threshold_bools, 'cluster'] = cluster_assignments
    return facedata_df

def cluster_facedata(
    face_data_across_frames,
    threshold,
    n_clusters=2
):
    """
    ---------------------------------------------------------------------------------------------------

    Clusters the face data across frames using embeddings.

    Parameters:
    ............
        face_data_across_frames (list): A list of face data across frames.
        threshold (float): The confidence threshold for clustering.
        n_clusters (int, optional): The number of clusters to create. Defaults to 2.

    Returns:
    ............
        pandas.DataFrame: The clustered face data.
    ---------------------------------------------------------------------------------------------------

    """
    facedata_df = facedata_list_to_df(
        face_data_across_frames
    )

    facedata_df = cluster_embeddings(
        facedata_df,
        confidence_threshold=threshold,
        n_clusters=n_clusters
    )

    return facedata_df

def create_single_face_output(
        cluster_df: pd.DataFrame, 
        min_frames_face_present: int,
        frames_per_row: int,
        fps: int,
        bbox_cols: list,
        interpolate=True
    ):
    """
    ---------------------------------------------------------------------------------------------------

    Identifies the frame indices in clusters where the number of frames is below a specified minimum threshold.

    Parameters:
    ............
    cluster_df : pandas.DataFrame
        DataFrame containing face cluster information with columns 'cluster_presences' and 'frame_idx'.
    min_frames_face_present : int
        Minimum number of frames required for a face to be considered present.
    frames_per_row : int
        Number of frames per row.
    fps : int
        Frames per second of the video.
    bbox_cols : list
        List of column names for the bounding box data.
    interpolate : bool
        Whether to interpolate missing values in the bounding box data.

    Returns:
    ............
    frame_idx_not_meeting_min : list
        List of frame indices that do not meet the minimum threshold.
    ---------------------------------------------------------------------------------------------------
    """
    interpolated_section_dfs = []
    cluster_df['cluster_presences'] = np.cumsum(
        cluster_df.sample_time.diff() > ((frames_per_row / fps)*1.5) # 1.5 is hacky but just gives rounding room for fps and frames per row
    )

    for _, presence_df in cluster_df.groupby('cluster_presences'):
        n_frames_present = len(presence_df) * frames_per_row
        if n_frames_present > min_frames_face_present:

            max_frame = presence_df.frame_idx.max()
            min_frame = presence_df.frame_idx.min()
            frames_in_clusters = list(range(min_frame, max_frame))
            upsampled_df = pd.DataFrame(frames_in_clusters, columns=['frame_idx'])
            bb_dict_df = presence_df[bbox_cols]

            merged_bb_df = upsampled_df.merge(
                bb_dict_df,
                how='outer',
                on='frame_idx'
            )
            if interpolate:
                interpolated_df = merged_bb_df.interpolate()
            else:
                interpolated_df = merged_bb_df.ffill()

            interpolated_section_dfs.append(interpolated_df)
    
    if len(interpolated_section_dfs) == 0:
        return pd.DataFrame(columns=bbox_cols)
    
    df_for_all_presences = pd.concat(interpolated_section_dfs)

    return df_for_all_presences


def prep_face_clusters_for_output(
    facedata_df, 
    min_frames_face_present, 
    capture_n_frames_per_second,
    fps,
    num_frames_vid,
    n_clusters,
    bbox_cols,
    interpolate=True
):
    """
    ---------------------------------------------------------------------------------------------------

    Filters and prepares face clusters for output.

    Parameters:
    ............
        facedata_df (pandas.DataFrame): The DataFrame containing face data.
        min_frames_face_present (int): The minimum number of frames a face must be present in a cluster.
        frames_per_row (int): The number of frames per row.
        fps (int): The frames per second of the video.
        num_frames_vid (int): The total number of frames in the video.
        n_clusters (int): The number of clusters.
        bbox_cols (list): The list of column names for the bounding box data.
        interpolate (bool): Whether to interpolate missing values in the bounding box data.

    Returns:
    ............
        dict: A dictionary containing face bounding box lists for each cluster.
    ---------------------------------------------------------------------------------------------------

    """
    
    facedata_df['sample_time']=facedata_df['frame_idx']/fps
    frames_per_row = round(fps/capture_n_frames_per_second)

    face_list_dict = {}
    for cluster_idx in range(n_clusters):
        out_df = pd.DataFrame(range(num_frames_vid),columns=['frame_idx'])
        cluster_df = facedata_df.loc[
            facedata_df.cluster==cluster_idx
        ]
        
        face_bbox_df = create_single_face_output(
            cluster_df,
            min_frames_face_present,
            frames_per_row,
            fps,
            bbox_cols,
            interpolate=interpolate
        )
        
        out_df = out_df.merge(
            face_bbox_df,
            how='outer',
            on='frame_idx'
        )

        bbox_df = out_df[bbox_cols].applymap(
            lambda x:int(x) if not pd.isna(x) else x
        )

        face_bbox_list = bbox_df.to_dict(orient='records')

        face_list_dict[cluster_idx] = face_bbox_list

    return face_list_dict

def preprocess_face_video(
    video_path,  
    n_people=2,
    capture_n_frames_per_second=3,
    model_name='Facenet',
    detector_backend='mtcnn', 
    face_threshold=.95, 
    min_sec_face_present=3, 
    n_frames=np.inf,
    interpolate=True
):
    """
    Preprocesses a face video by extracting face data, clustering the faces, and preparing the output.

    Parameters:
    - video_path (str): Path to the input video file.
    - n_people (int): Number of people in video used to set number of clusters to create. Default is 2, assuming this is a clinical interview.
    - capture_n_frames_per_second (int): Number of frames to capture per second. Increasing this parameter should increase the quality of clustering but will cause the function to run for longer. Default is 2.
    - model_name (str): Name of the face recognition model to use. Default is 'Facenet'.
    - detector_backend (str): Backend to use for face detection. Can be any model used in the Deepface library. Default is 'mtcnn'.
    - face_threshold (float): Similarity threshold for identifying faces. Default is 0.95.
    - min_sec_face_present (int): Minimum number of seconds a face must be present after clustering to not be filtered out. Default is 3.
    - n_frames (int): Maximum number of frames to process. Default is np.inf (i.e. process all frames).
    - interpolate (bool): Whether to interpolate missing values in the bounding box data. Default is True.
   
    Returns:
    - bb_dict (dict): Dictionary containing the framewise bounding boxes for each face of n_people in video. keys are zero indexed n_people integers.
    - facedata_df (pandas.DataFrame): DataFrame containing the face data and cluster info.
    """

    config = get_config(os.path.abspath(__file__), 'preprocess.json')

    bb_dict = {}
    facedata_df = pd.DataFrame()
    bbox_cols=config["bbox_dict_cols"]

    try:
        face_data_across_frames, fps, num_frames_vid = load_facedata_from_video(
            video_path,
            config['rgb_back_ends'],
            n_frames=n_frames,
            capture_n_per_sec=capture_n_frames_per_second,
            detector_backend=detector_backend,
            model_name=model_name
        )

        facedata_df = cluster_facedata(
            face_data_across_frames,
            face_threshold,
            n_clusters=n_people
        )

        min_frames_face_present = fps * min_sec_face_present
        bb_dict = prep_face_clusters_for_output(
            facedata_df, 
            min_frames_face_present, 
            capture_n_frames_per_second,
            fps,
            num_frames_vid,
            n_people,
            bbox_cols=bbox_cols,
            interpolate=interpolate
        )

    except Exception as e:
        logger.info(f"Error preprocessing video: file: {video_path} & Error: {e}'")
        
    return  bb_dict, facedata_df



import cv2
import numpy as np
import logging

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

def crop_img(img, bb_dict):
    """
    ---------------------------------------------------------------------------------------------------

    This function takes an RGB image and a dictionary containing bounding box coordinates as input. 
    It crops the image based on the provided bounding box coordinates and returns the cropped region.

    Parameters:
    ----------
    img : numpy.ndarray
        The RGB image to be cropped.
    bb_dict : dict
        A dictionary containing the bounding box coordinates.
        The dictionary should have the following keys:
            - 'bb_x': The x-coordinate of the top-left corner of the bounding box.
            - 'bb_y': The y-coordinate of the top-left corner of the bounding box.
            - 'bb_w': The width of the bounding box.
            - 'bb_h': The height of the bounding box.

    Returns:
    -------
    numpy.ndarray
        The cropped region of the image.

    ---------------------------------------------------------------------------------------------------
    """

    x = int(bb_dict['bb_x'])
    y = int(bb_dict['bb_y'])
    w = int(bb_dict['bb_w'])
    h = int(bb_dict['bb_h'])
    roi = img[y:y+h, x:x+w]
    return roi

def draw_bounding_boxes_sf(frame, bb_dict):
    """
    ---------------------------------------------------------------------------------------------------

    Draw bounding boxes on the given frame.

    Parameters:
    ............
        frame (numpy.ndarray): The frame on which to draw the bounding boxes.
        bb_dict (dict): A dictionary containing the bounding box coordinates.

    Returns:
    ............
        numpy.ndarray: The frame with the bounding boxes drawn on it.
    ---------------------------------------------------------------------------------------------------
    
    """
    x = int(bb_dict['bb_x'])
    y = int(bb_dict['bb_y'])
    w = int(bb_dict['bb_w'])
    h = int(bb_dict['bb_h'])
    cv2.rectangle(frame, (x, y), (x+w, y+h), (0, 255, 0), 2)
    return frame

def calculate_padding(bb_dict, padding_percent):
    """
    Calculate the padding around the bounding box based on a percentage.

    Parameters:
    ----------
    bb_dict : dict
        A dictionary containing the bounding box coordinates.
        The dictionary should have the following keys:
            - 'bb_x': The x-coordinate of the top-left corner of the bounding box.
            - 'bb_y': The y-coordinate of the top-left corner of the bounding box.
            - 'bb_w': The width of the bounding box.
            - 'bb_h': The height of the bounding box.
    padding_percent : float
        The percentage of padding to add around the bounding box.

    Returns:
    -------
    dict
        A dictionary containing the new bounding box coordinates with padding.
    """
    x = int(bb_dict['bb_x'])
    y = int(bb_dict['bb_y'])
    w = int(bb_dict['bb_w'])
    h = int(bb_dict['bb_h'])
    
    padding_x = int(w * padding_percent)
    padding_y = int(h * padding_percent)
    
    return {
        'bb_x': max(0, x - padding_x),
        'bb_y': max(0, y - padding_y),
        'bb_w': w + 2 * padding_x,
        'bb_h': h + 2 * padding_y
    }

def resize_to_fit(img, frame_size):
    """
    Resize the image to fit within the frame size while maintaining the aspect ratio.

    Parameters:
    ----------
    img : numpy.ndarray
        The image to be resized.
    frame_size : tuple
        The size of the frame (width, height).

    Returns:
    -------
    numpy.ndarray
        The resized image that fits within the specified frame size.
    """
    h, w = img.shape[:2]
    frame_w, frame_h = frame_size
    
    aspect_ratio = w / h
    frame_aspect_ratio = frame_w / frame_h
    
    if aspect_ratio > frame_aspect_ratio:
        # Width is the limiting factor
        new_w = frame_w
        new_h = int(new_w / aspect_ratio)
    else:
        # Height is the limiting factor
        new_h = frame_h
        new_w = int(new_h * aspect_ratio)
    
    resized_img = cv2.resize(img, (new_w, new_h), interpolation=cv2.INTER_AREA)
    return resized_img

def center_in_frame(cropped_img, frame_size=(512, 512), background_color=(0, 0, 0)):
    """
    Center the cropped and padded image in a specified frame.

    Parameters:
    ----------
    cropped_img : numpy.ndarray
        The cropped image to be centered.
    frame_size : tuple, optional
        The size of the frame (default is 512x512 pixels).
    background_color : tuple, optional
        The background color of the frame (default is black, (0, 0, 0)).

    Returns:
    -------
    numpy.ndarray
        The image centered within the specified frame.
    """
    h_padded, w_padded = cropped_img.shape[:2]
    
    if w_padded > frame_size[0] or h_padded > frame_size[1]:
        cropped_img = resize_to_fit(cropped_img, frame_size)
        h_padded, w_padded = cropped_img.shape[:2]
    
    frame = np.full((frame_size[1], frame_size[0], 3), background_color, dtype=np.uint8)
    
    x_offset = (frame_size[0] - w_padded) // 2
    y_offset = (frame_size[1] - h_padded) // 2
    
    frame[y_offset:y_offset + h_padded, x_offset:x_offset + w_padded] = cropped_img
    
    return frame

def crop_with_padding_and_center(
    img, 
    bb_dict, 
    padding_percent=0.1, 
    frame_size=(512, 512),
    background_color=(0, 0, 0)
):
    """
    Crop an image with padding around the bounding box and center it in a frame.

    Parameters:
    ----------
    img : numpy.ndarray
        The RGB image to be cropped and centered.
    bb_dict : dict
        A dictionary containing the bounding box coordinates.
    padding_percent : float, optional
        The percentage of padding to add around the bounding box (default is 10%).
    frame_size : tuple, optional
        The size of the frame (default is 512x512 pixels).
    background_color : tuple, optional
        The background color of the frame (default is black, (0, 0, 0)).

    Returns:
    -------
    numpy.ndarray
        The padded and centered image within the specified frame.
    """
    padded_bb_dict = calculate_padding(bb_dict, padding_percent)
    padded_crop = crop_img(img, padded_bb_dict)
    final_img = center_in_frame(padded_crop, frame_size, background_color)
    
    return final_img

def blacken_outside_bounding_box(frame, bb_dict):
    """
    ---------------------------------------------------------------------------------------------------

    Blackens the area outside a specified bounding box in the given frame.

    Parameters:
    ............
    frame : numpy.ndarray
        The input image frame in which the area outside the bounding box will be blackened.
    bb_dict : dict
        Dictionary containing bounding box coordinates and dimensions with keys:
        'bb_x' (int): The x-coordinate of the top-left corner of the bounding box.
        'bb_y' (int): The y-coordinate of the top-left corner of the bounding box.
        'bb_w' (int): The width of the bounding box.
        'bb_h' (int): The height of the bounding box.

    Returns:
    ............
    numpy.ndarray
        The resulting image frame with the area outside the bounding box blackened.
    ---------------------------------------------------------------------------------------------------

    """
    x = int(bb_dict['bb_x'])
    y = int(bb_dict['bb_y'])
    w = int(bb_dict['bb_w'])
    h = int(bb_dict['bb_h'])

    # Create a black image of the same size as the frame
    mask = np.zeros_like(frame)

    # Copy the bounding box area from the original frame to the mask
    mask[y:y+h, x:x+w] = frame[y:y+h, x:x+w]

    return mask

def create_cropped_frame(
    frame,
    frame_dict,
    crop=True,
    default_size=(512,512),
    ):
    """
    Blackens the area outside a specified bounding box in the given frame or crops the frame with padding and centering.

    Parameters:
    ----------
    frame : numpy.ndarray
        The input image frame in which the area outside the bounding box will be blackened or cropped.
    frame_dict : dict
        Dictionary containing bounding box coordinates and dimensions with keys:
        - 'bb_x' (int): The x-coordinate of the top-left corner of the bounding box.
        - 'bb_y' (int): The y-coordinate of the top-left corner of the bounding box.
        - 'bb_w' (int): The width of the bounding box.
        - 'bb_h' (int): The height of the bounding box.
    crop : bool, optional
        If False, blackens the area outside the bounding box. If True, crops the frame with padding and centering. Default is True.
    default_size : tuple, optional
        The size of the cropped frame if `crop` is True. Default is (512, 512).

    Returns:
    -------
    numpy.ndarray
        The resulting image frame with the area outside the bounding box blackened or the cropped frame.

    Notes:
    ------
    - If `crop` is False, the function uses the `blacken_outside_bounding_box` function to blacken the area outside the bounding box.
    - If `crop` is True, the function uses the `crop_with_padding_and_center` function to crop the frame with padding and centering.
    """
    
    if crop:
        face_frame = crop_with_padding_and_center(frame, frame_dict, frame_size=default_size)
    else:
        face_frame = blacken_outside_bounding_box(frame, frame_dict)
        
    return face_frame


def create_face_frame(
        frame_dict,
        frame,
        default_size=(512,512),
        trim=True,
        debug=False,
        crop=True
        ):
    """
    Create a face frame by cropping the input frame based on the provided frame dictionary and booleans for
    style of cropping.

    Parameters:
    ----------
        frame_dict (dict): A dictionary containing the coordinates of the face region in the frame.
        frame (numpy.ndarray): The input frame to be cropped.
        default_size (tuple, optional): The default size of the cropped face frame. Defaults to (512, 512).
        trim (bool, optional): Whether to the trim video so only frames with face present are retained. Defaults to True.
        crop (bool, optional): Whether to crop the cropped face frame. Defaults to True.

    Returns:
        numpy.ndarray: The cropped face frame.
    """
    
    if (len(frame_dict.keys()) != 0) and not np.isnan(frame_dict['bb_x']):

        face_frame = create_cropped_frame(
            frame,
            frame_dict,
            crop=crop,
            default_size=default_size,
        )

    elif debug:

        face_frame = frame

    elif not trim:
        
        if crop:
            face_frame = np.zeros(default_size+(3,),dtype=np.uint8)     
        else:
            face_frame = np.zeros_like(frame,dtype=np.uint8)
            
    else:

        face_frame = np.array([])
    
    return face_frame

def create_cropped_video(
    video_path,
    detections,
    output_path,
    crop=False,
    trim=True,
    debug=False,
    default_size_for_cropped=(512, 512)
):
    """
    Process the video by drawing bounding boxes based on the provided detections
    and save the processed video as an mp4.

    Parameters:
    ----------
    video_path : str
        Path to the video file.
    detections : list
        List of dictionaries containing bounding box and frame index information.
    output_path : str
        Path to save the output video.
    crop : bool, optional
        Flag to crop frame so that only bounding box is retained versus setting all pixels outside bounxing box to black but retaining original framesize (crop == False). by default False.
    trim : bool, optional
        Flag to only return frames with bounding box information, effectively altering the length of the video by. If false black frames are appended when no bounding box info is present default True
    debug : bool, optional
        Flag to keep the original frames when no face is detected, by default False.
    default_size_for_cropped : tuple, optional
        Default size for the cropped frames, by default (512, 512).

    Raises:
    ------
    ValueError
        If there is an error opening the video file.

    """
    cap = cv2.VideoCapture(video_path)

    if debug:
        if crop:
            logger.warning("Debug mode is enabled. crop flag will be set to False.")
            crop = False


    if not cap.isOpened():
        raise ValueError(f"Error opening video file: {video_path}")

    # Get video properties
    fps = cap.get(cv2.CAP_PROP_FPS)

    frame_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

    if crop:
        frame_width, frame_height = default_size_for_cropped

    # Initialize the video writer
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')  # or 'XVID', 'MJPG', etc.
    out = cv2.VideoWriter(output_path, fourcc, fps, (frame_width, frame_height))

    frame_index = 0
    max_frame_index = len(detections)

    while True:
        ret, frame = cap.read()
        if not ret:
            break
        if frame_index >= max_frame_index:
            break

        frame_dict = detections[frame_index]

        face_frame = create_face_frame(
            frame_dict,
            frame,
            default_size=default_size_for_cropped,
            trim=trim,
            debug=debug,
            crop=crop
        )

        # Check if face_frame is not empty (i.e., face detected)
        if face_frame.size != 0:
            out.write(face_frame)

        frame_index += 1

    cap.release()
    out.release()


from .face_landmark import (
    facial_expressivity,
)
from .facial_emotion import (
    emotional_expressivity,
)

from .eye_blink import (
    eye_blink_rate,
)

from .preprocess_video import (
    preprocess_face_video
)

from .util import (
    create_cropped_video
)

__all__ = ["facial_expressivity", "emotional_expressivity", "eye_blink_rate", "preprocess_face_video", "create_cropped_video"]



# author:    Kieran McVeigh
# website:   http://www.bklynhlth.com
from sklearn.mixture import GaussianMixture
import pandas as pd

def get_summary(df, start_at_col):
    """
    ---------------------------------------------------------------------------------------------------

    This function calculates the summary measurements from the framewise displacement data.

    Parameters:
    ............
    df : pandas.DataFrame
        framewise euclidean displacement dataframe
    start_at_col : int
        column index to start summary calculations

    Returns:
    ............
    df_summ : pandas.DataFrame
         stat summary dataframe

    ---------------------------------------------------------------------------------------------------
    """

    df_summ = pd.DataFrame()
    if len(df.columns)>0:
        df_mean = pd.DataFrame(df.mean()).T.iloc[:, start_at_col:].add_suffix('_mean')
        df_std = pd.DataFrame(df.std()).T.iloc[:, start_at_col:].add_suffix('_std')

        df_summ = pd.concat([df_mean, df_std], axis =1).reset_index(drop=True)
    return df_summ

def get_fps(df):
    """
    ---------------------------------------------------------------------------------------------------
    Calculate the frames per second (FPS) from a DataFrame.

    This function computes the FPS by taking the reciprocal of the mode of the time differences between consecutive rows in the DataFrame.

    Parameters:
    df : DataFrame
        A DataFrame containing a 'time' column with timestamps.

    Returns:
    int: The calculated frames per second (FPS).
    ---------------------------------------------------------------------------------------------------
    """
    #check if df.time exists
    if 'time' not in df.columns:
        raise ValueError('DataFrame must contain a time column.')
    return int(1/df.time.diff().mode())

def get_speaking_probabilities(df, rolling_std_seconds):
    """
    ---------------------------------------------------------------------------------------------------
    Calculate the probability of speaking at each frame in a DataFrame.

    This function calculates the probability of speaking at each frame in a DataFrame by fitting a Gaussian Mixture Model to the rolling standard deviation of the 'mouth_openness' column.

    Parameters:
    df : pd.DataFrame
        A DataFrame containing a 'time' column with timestamps and a 'mouth_openness' column with mouth openness values.
    rolling_std_seconds : int
        The number of seconds over which to calculate the rolling standard deviation.

    Returns:
    pandas.Series: A Series containing the probability of speaking at each frame.
    ---------------------------------------------------------------------------------------------------
    """
    fps = get_fps(df)
    rolling_std_frames = int(rolling_std_seconds*fps)
    df = df.copy(deep=True)
    df['rolling_mouth_open_std'] = df.mouth_openness.rolling(
        rolling_std_frames,
        min_periods=2
    ).std()


    df_nona = df[['frame','time','rolling_mouth_open_std']].dropna()
    gmm = GaussianMixture(n_components=2)
    gmm.fit(df_nona.rolling_mouth_open_std.values.reshape(-1,1))
    prob_preds = gmm.predict_proba(df_nona.rolling_mouth_open_std.values.reshape(-1,1))
    mean_1, mean_2 = gmm.means_.flatten()
    if mean_1 > mean_2:
        df_nona['speaking'] = prob_preds[:,0]
    else:
        df_nona['speaking'] = prob_preds[:,1]
        
    df = df.merge(df_nona, on='frame', how='left')
    return df.speaking

def split_speaking_df(df_disp, speaking_col, start_at_col):
    """
    ---------------------------------------------------------------------------------------------------

    This function splits the displacement dataframe into two dataframes based on speaking probability.

    Parameters:
    ............
    df_disp : pandas.DataFrame
        displacement dataframe
    speaking_col : str
        speaking probability column name
    start_at_col : int
        column index to start summary calculations


    Returns:
    ............
    df_summ : pandas.DataFrame
        stat summary dataframe
    ---------------------------------------------------------------------------------------------------
    """
    speaking_df = df_disp[df_disp[speaking_col] > 0.5]
    not_speaking_df = df_disp[df_disp[speaking_col] <= 0.5]
    speaking_df = speaking_df.drop(speaking_col, axis=1)
    not_speaking_df = not_speaking_df.drop(speaking_col, axis=1)

    speaking_df_summ = get_summary(speaking_df, start_at_col)
    not_speaking_df_summ = get_summary(not_speaking_df, start_at_col)
    speaking_df_summ = speaking_df_summ.add_suffix('_speaking')
    not_speaking_df_summ = not_speaking_df_summ.add_suffix('_not_speaking')
    
    df_summ = pd.concat([speaking_df_summ, not_speaking_df_summ], axis=1)
    
    return df_summ
